from pathlib import Path

__all__ = [
    'NEUROLIB_CACHE_DIRECTORY',
    'NEUROLIB_DATASET_DIRECTORY',
    'ATLAS_CACHE_DIRECTORY',
    'CASCADE_MODEL_CACHE_DIRECTORY'
]

CACHE_DIRECTORY = Path.home() / '.cache'
NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'
NEUROLIB_DATASET_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'dataset'

# ATLAS
ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'

# MODEL
CASCADE_MODEL_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'cascade'
