"""
Locomotion
============

This module provide usage for selected locomotion epoch
Including running/stationary, remove jumping frames, interpolation of gaps...


"""

from .epoch import *
from .position import *
